/*
  what it does:
   measure speed of mpi-transfer between tasks (more than just 2) and
   switching source and destination of transfer at every cycle (thats why
   it is stressing the network) or simply do MPI_alltoall

  linux:
    mpicc -O2 -lm -o mpi_stress mpi_stress.c
    mpicc -O2 -lm -D CFG_CPUSET=0x00ff -o mpi_stress mpi_stress.c
    # cpuset is a bit pattern (hex) for available cpu cores per node
    #   if you want bind tasks to cores using sched_setaffinity()
    mpirun -np 8 ./mpi_stress --bsize 26 | tee mpi_stress.log
    # example with tcp + tcp_over_ib
    mpiexec --mca btl self,sm,tcp --mca btl_tcp_if_include eth0 --mca
     btl_tcp_if_include ib0 --hostfile $PBS_NODEFILE --bynode -np 8 numactl
     --localalloc --physcpubind 0 ./mpi_stress --bsize 30 --alltoall 
   
  see mpi_stress.gpl and at bottom of this file for some data!
  
  Changelog:
    2014-07-27 - add --minfree=.. to check vmem needed for mpi-calls
    2014-07-01 - cpu-pinning per option
    2014-06-19 - add option --bsize0 log2_min_block_size (debugging)
    2014-06-16 - add min loop check (loops >= (mpi_n-1)) for accuracy
    2013-06-21 - more compact output, aBW=SumBW (a=aggregated), t/s=time[s]
    2011-10-24 - better timing error for A2A too + memory constraints
    2011-10-20 - limit msg size to 1GB (MPI restriction for MPI_CHAR)
               - blocksize from min to max (old: max to min)
               - compute error by time variance and adapt runtime
               - check for bit errors
    2008-08-23 - change MPI_Sendrecv loop to avoid bad measurement on NUMA
                 systems (Altix4700: 8-9GB/s measured, but 1.3GB/s real) 
    2008-08-20 - fix race condition, have lead
                 to speed-down-jumps, MPI_ERR_TRUNCATE and dead-locks at end
                 (thanks to David Gingold)
               - better option handling
  Problems:
    - one loop is only part of a full all2all, where distribution
      of remote and local connection and speed results depend from
      the distribution, getting good results after (threads/node-1)
      cycles but that would be takes long for big Nodes! (2011-10)
      ToDo: successiv searching slowest and fastest N/2-pairing ???
        i searches i+1..i-1 for slowest connect, after that next
        unpaired node does the same until all nodes paired (N steps)
        2nd run same with fastest pairing (all local! not useful?)

  ToDo:
    - alloc asize * bsize block z.B. 2048 * 1MB = 2GB/task
      then transfer node[i  ].mem1[block*(j%asize)]
                 to node[i+j].mem2[block*(j%asize)]
            more realistic
      old was: transfer node[i  ].mem1[0] to node[i+j].mem2[0]
           arteficial full mem transfers
      better: v1 += sparseH * v2 # using fixed nz/line
    - async mpi_stress + cpu for max BW to test old elkos of DC-DC converter
      s. mkautz 2014-06
    - CPU_SET like memspeed.c (break 32 cores limit, or slurm cpu pinning)
    - use local mem + transfer mem (do async memset|memcopy or similar
      to avoid thread jumping in SMPs, which generate arteficial
      transfer speeds)  do we understand that?
    - old timings?
    - send more random bit pattern (checks the memory correctness)
    - add initial 2 node latency and speed (able to compute theoretical max)
      compute min vs. max of pair speed (detect local + remote pairs)
    - SMP influence of cache coherence (disable?)
    - add a all send to one OOM-pressure test (s. vector.c mpich failes)
    - check 0-to-i homogenity add beginning (net topology)

  FAQ:
   - Q: Why not MPI_alltoall? Could better optimize for topology.
     A: Must have much (N times) more memory for buffering transmitted data.
        In real app we can transfer parts, compute and free the buffer.

*/

#define _GNU_SOURCE 1 // this defines __USE_GNU in features.h for sched.h

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h> 
// #include <time.h>
#include <malloc.h> /* memalign() */
#include <string.h>
#include <math.h>  /* fabs() */

#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
# include <unistd.h>  // declare sleep()
# include <sched.h>   // declare sched_setaffinity()
# ifndef _FEATURES_H  
#  warning "_FEATURES_H is not defined, try top add #include <features.h>"
# endif
# ifndef __USE_GNU                                                 
#  warning "expect __USE_GNU defined, but it isn't, try -D_GNU_SOURCE=1"
# endif
#endif

int mpi_n, mpi_i,
     node_dst,  node_src, /* destination and source node */
    *next_dst, *next_src, /* next destination and source node*/
    mintime=1; /* minimum seconds run time (max. is 2*mintime) */
char *buf1, *buf2; /* transfer buffers */
static int header=0;

int test_sendrecv(unsigned long len){
  unsigned long i1, i2, i3;
  double t1, t2, t3=0;
  MPI_Status status;
  if (!header) { header=1;
    if (!mpi_i)
    printf("# tasks msgsize/B lg2loops t/s"
           " latency/s aBW[B/s] BW[B/s] err%%\n"); /* BW=bandwidth*/
  }
  i3=1;  // log2 of minloops
  // ToDo: dont double on timing errors, just repeat and print min/max time
  //   this would be better on switches which perform poor at critical usage
  //  t>mintime abbrechen (done), min/maxspeed ab mintime/2^4 ???
  // ToDo: 2014-06 better i1=(mpi_n-1)<<i3 for full cycle?
  //      bei kautz = 5700 (2^12..2^12) sehr viel!? ToDo
  t2=0.0; t1=MPI_Wtime();
  for (i2=0,i1=1l<<i3;i1/*stop at int overflow*/;i1<<=1,i3++){
    // i1 = log2 of next inner loops before next time measurement
    // i2 = loops from start time t1
    for (;2*i2<i1;i2++) {  /* include shorter runs to spare usertime */
      /* sendcount is integer 32bit, max block is 2GB! */
      MPI_Sendrecv(buf1,len,MPI_BYTE,node_dst,0,
                   buf2,len,MPI_BYTE,node_src,0,MPI_COMM_WORLD,&status);
      node_src=next_src[node_src];
      node_dst=next_dst[node_dst];
      /* do a 2nd call to avoid wrong measurement on NUMA systems */
      MPI_Sendrecv(buf2,len,MPI_BYTE,node_dst,0,
                   buf1,len,MPI_BYTE,node_src,0,MPI_COMM_WORLD,&status);
      node_src=next_src[node_src];
      node_dst=next_dst[node_dst];
    }
    t3=t2; t2=MPI_Wtime()-t1;
    /* avoid a race condition, added time will vanish for bigger i3 */
    // ToDo: replace by min/max to get noshift+pairwise SMP/TCP ranges
    MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD); /* t.rank0 only */
    // t3+=MPI_Wtime()-t1-t2;
    /* t2 must be the same on all threads! */
// on bad networks compute time may increase dramatically 2013-10-18
//    if ((i2>1 && t2>20*(mpi_n-1)*mintime) /* maxtime for: N=16 1..300s */ ||
//      (fabs(2*t3-t2)*200<(t2+2*t3)  && t2>mintime) ) break;
    /* if i2 is below mpi_n-1 we dont have all paths and get inexact results
     *   example: on 2 smp-nodes 2*16 where local mpi is faster */
    if (t2>mintime && i2>=2*(mpi_n-1)) break;  /* 2014-06-13 better accur. */
    // ToDo: ... < (2*t3+t2)/2 see comp2 256 tasks
    if (t2>10 && !mpi_i) // some alive messages
    printf("# %4d %9lu %6lu %6.2f %8.2e  %8.2e %8.2e %6.2f\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len*mpi_n/t2, i1*1.*len/t2,
      (2*t3-t2)*100/t2);
    if (t2>mintime && i2<(mpi_n-1) && !mpi_i)
    printf("# cycle number still below thread number, continue...\n");
  }
  if (!mpi_i)
    printf("  %4d %9lu %6lu %6.2f %8.2e  %8.2e %8.2e %6.2f\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len*mpi_n/t2, i1*1.*len/t2,
      (2*t3-t2)*100/t2);
  return 0;
}

int test_alltoall(unsigned long len){
  unsigned long i1, i2, i3;
  double t1, t2, t3=0;
  if (!header) { header=1;
    if (!mpi_i)
    printf("# tasks msgsize/B lg2loops t/s"
           " latency/s aBW[B/s] BW[B/s] Err%%\n"); /* BW=bandwidth*/
  }
  i3=0;
  t1=MPI_Wtime();
  for (i2=0,i1=1l<<i3;i1;i1<<=1,i3++){
    for (;i2<i1;i2++) {  /* include shorter runs to spare usertime */
      MPI_Alltoall(buf1,len,MPI_BYTE,
                   buf2,len,MPI_BYTE,MPI_COMM_WORLD);
      // i=1..n: MPI_Send(buff+i*len*sizeof(MPI_BYTE),len,MPI_BYTE,i,...      
      // i=1..n: MPI_Recv(buff+i*len*sizeof(MPI_BYTE),len,MPI_BYTE,i,...      
    }
    t3=t2; t2=MPI_Wtime()-t1;
    /* avoid a race condition, added time will vanish for bigger i3 */
    MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    // t3+=MPI_Wtime()-t1-t2;
// on bad networks compute time may increase dramatically 2013-10-18
//    if ((i2>1 && t2 > 64*mintime) ||   /* hard end, saw problems on tcp GbE at 256B */
//      (fabs(2*t3-t2)*200<(t2+2*t3) && t2>mintime) ) break; /* 0.5% */
    if (t2>mintime) break;
    if (t2>10 && !mpi_i)
    printf("# %4d %9lu %6lu %6.2f %8.2e  %8.2e %8.2e %5.2f\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len*mpi_n*mpi_n/t2, 
      i1*1.*len*mpi_n/t2, (2*t3-t2)*200/(t2+2*t3));
      /* time-error:  2*oldrun(oldloops) vs. 1*newrun(2*oldloops) */
  }
  /* we show msgsize as single transfer message bow, 1.. */
  if (!mpi_i)
    printf("  %4d %9lu %6lu %6.2f %8.2e  %8.2e %8.2e %5.2f\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len*mpi_n*mpi_n/t2,
       i1*1.*len*mpi_n/t2, (2*t3-t2)*200/(t2+2*t3));
  return 0;
}


int main(int argn, char *argv[]) {
  unsigned long i1,
      minfree_kB=~0L, /* 2014-07-27 memeater reserve to test mpi-calls */
      bsize0=1, /* 2014-06-19 for debugging start with 1...bsize */
      bsize;  /* buffer size or message size, 2^bsize2 */
  int version, subversion, i, i2,
      pin_flag=0, /* thread to processor pinning 2014-07-01 */
      num_bflips=0, /* 2014-06 count byteflip-errors */
      bsize2=26, /* log2 of bsize = 64MB */
      pairs=0,  /* rings (src!=dst, default) versus pairs (src==dst) */
      nshift=1, /* node shift, 0=no (fixed src/dst), 1=yes (shifting src/dst) */
      ata=0,    /* MPI_Sendrecv (default) versus MPI_Alltoall */
      align=1;  /* align memory to 64 (default) or not */
  setvbuf(stdout,NULL,_IONBF,0);
  if (MPI_Init(&argn, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  MPI_Get_version(&version, &subversion);
  if (!mpi_i)
    printf("# mpi_stress v2014-07-27 (timing: v2011-10-24)\n");
  for (i=1;i<argn;i++) {
    if (i+1<argn && !strcmp(argv[i],"--bsize0")) { /* start blocksize */
            i++; bsize0=1l<<atol(argv[i]); continue; }
    if (i+1<argn && !strcmp(argv[i],"--bsize")) {
            i++; bsize2=atol(argv[i]); continue; }
    if (i  <argn && !strncmp(argv[i],"--bsize0=",9)) { /* 2014-07 add opt= */
            bsize0=1l<<atol(argv[i]+9); continue; }
    if (i  <argn && !strncmp(argv[i],"--bsize=",8)) {
            bsize2=atol(argv[i]+8); continue; }
    if (i  <argn && !strncmp(argv[i],"--mintime=",10)) {
            mintime=atol(argv[i]+10); continue; }
    if (i+1<argn && !strcmp(argv[i],"--mintime")) {
            i++; mintime =atoi(argv[i]); continue; }
    if (i  <argn && !strncmp(argv[i],"--minfree=",10)) {
            minfree_kB =atoi(argv[i]+10); continue; }
    if (!strcmp(argv[i],"--noshift")) { nshift=0; continue; }
    if (!strcmp(argv[i],"--alltoall")) { ata=1;   continue; }
    if (!strcmp(argv[i],"--pairwise")) { pairs=1; continue; }
    if (!strcmp(argv[i],"--noalign"))  { align=0; continue; }
    /* pin_* = processor core/thread pinning */
    if (strncmp(argv[i],"--pin_ncpu=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i],"--pin_ofs=", 10)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i],"--pin_dist=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i],"--pin_size=",11)==0) { pin_flag=1; continue; }
    if (!mpi_i)
      printf("valid options: [--bsize0=<log2minsize>] [--bsize=<log2maxsize>]\n"
             "  [--mintime=<seconds>] [--minfree=<kB_left_by_memeater>]\n"
             "  [--alltoall|{[--pairwise] [--noshift]}] [--noalign]\n"
             "  [--pin_ncpu=16 --pin_ofs=0 --pin_dist=1 --pin_size=1]\n"
     );
  }
  /* Warning added JS2011-10, msgsize is int32 */
  if ((ata==0 && bsize2>30)
   || (ata==1 && 1l<<bsize2 > mpi_n*1l<<30)) {
    if(!mpi_i)
      printf("# WARNING: MPI can handle 2^30 Byte-blocks (1GB) at maximum\n");
    /* ToDo: memory size and blocks differ in case of alltoall, mem/task? */
    if (ata==0) bsize2=30;
    else for ( ; 1l<<bsize2 > mpi_n*1l<<30; bsize2--); // set to maximum
  }
  bsize = 1l << bsize2;
  if (bsize0<    1) bsize0=1;
  if (bsize0>bsize) bsize0=bsize;

#if !defined(CFG_CPUSET)
  if (pin_flag==1 && !mpi_i)
    printf("WARNING: mpi_stress compiled without CFG_CPUSET\n");
#endif
#if defined(CFG_CPUSET) && defined(__USE_GNU)
  if (pin_flag==1) { /* processor/thread pinning */
   int j1;
   // ToDo: openmp code missing
   // this processor bounding is useful for the case, that the jobsystem
   //   does not do this automatically
   // CFG_CPUSET is the bitmap for available SMP-CPUs per node
   // CFG_CPUSET=0x0005: use core 0 and 2 (Nodes with 2 Hyper-Threading CPUs)
   //   ToDo: define on runtime per taskset + sched_getaffinity ? (jobsystem?)
   // use --byslot for numbits(CFG_CPUSET)<mpi_n
   cpu_set_t cpuset; // typedef struct { ULong bits[CPU_SETSIZE / NCPUBITS]; }
   // new CPU pinning method for more than 32 cores
   struct s_pin {int ncpu;int ofs;int dist;int size;}; // new2013-05-07
   // most important value first (numcpus, then offset, then distance,...)
   int  ncores=0, ii; unsigned long cmask=0L;
   struct s_pin pin = {16,0,1,1}; // CFG_CPUSET=16,0,1,1 mask=0xffff
   // struct s_pin pin = {8,1,4,2}; // mask=0x00066
   for (i1=1;i1<argn;i1++) { /* check all args */  
      if (strncmp(argv[i1],"--pin_ncpu=",11)==0) { 
         pin.ncpu=atoi(argv[i1]+11); continue; }   
      if (strncmp(argv[i1],"--pin_ofs=",10)==0) {  
         pin.ofs= atoi(argv[i1]+10); continue; }   
      if (strncmp(argv[i1],"--pin_dist=",11)==0) { 
         pin.dist=atoi(argv[i1]+11); continue; }   
      if (strncmp(argv[i1],"--pin_size=",11)==0) { 
         pin.size=atoi(argv[i1]+11); continue; }   
   }

   if (pin.ncpu<1) {printf("ERROR: pin.ncpu<1\n");exit(1);}
   if (pin.size<1) {printf("ERROR: pin.size<1\n");exit(1);}
   if (pin.dist<0) {printf("ERROR: pin.dist<1\n");exit(1);}
   if (pin.ofs>=pin.ncpu) {printf("ERROR: pin.shift\n");exit(1);}

   // auto-detect num cores by try+fail sched_affinity
   CPU_ZERO(&cpuset); 
   for(j1=0,i=CPU_SETSIZE;i;i>>=1){
     CPU_SET(j1|i,&cpuset);
     if (!sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) j1|=i; // no_err!
     CPU_CLR(j1|i,&cpuset);
   }
   if (!mpi_i) printf("# Max-detected-CPU-Cores/Threads= %d\n", j1+1);
   ncores=j1+1;
      
   for (ii=0;ii<8*sizeof(long) && ii<mpi_n;ii++) {  // set cmask
     j1=((ii/pin.size)*pin.dist
        +(ii%pin.size)+pin.ofs)%pin.ncpu;
     if (j1<8*sizeof(long)) cmask |= 1UL<<j1;
   }
   if (!mpi_i)
     printf("# CPU-Pinning: idx=((i/%d)*%d+(i%%%d)+%d)%%%d, mask=0x...%08lx\n",
     pin.size, pin.dist, pin.size, pin.ofs, pin.ncpu,
      cmask);
   if (pin.ncpu>ncores && !mpi_i) printf("WARNING: pin.ncpu>ncores\n");

   // does only cover equidistant distribution but most
   // MASK     
   // 0xffff = {16,1,1,0}
   // 0x6666 = {16,2,4,1}
   // 0x0f0f = {16,4,8,0} = ((ii/4)*8+0)%16
   // 0xf0f0 = {16,4,8,4}
   ii=((mpi_i/pin.size)*pin.dist
      +(mpi_i%pin.size)+pin.ofs)%pin.ncpu;
   // ncores=pin.ncpu;
   if (ii>=CPU_SETSIZE) {printf("ERROR: ii>CPU_SETSIZE, ii=%d\n",ii);exit(1);}
   CPU_ZERO(&cpuset);   // void CPU_ZERO(cpu_set_t *set); bits[idx]|=mask
   CPU_SET(ii,&cpuset); //  ->bits[ii/NCPUBITS] |= (1<<(ii%NCPUBITS))
   
   // The constant CPU_SETSIZE (1024) specifies a value one greater than  the
   //      maximum CPU number that can be stored in a CPU set.
   usleep((mpi_i%16+1)*10000L); // 10ms..170ms, sort outputs  
   if (mpi_i<4 || mpi_i>=mpi_n-2) {
     printf("# set cpu affinity for core(%d)=%d size=%d\n",
              mpi_i, ii, CPU_SETSIZE);
   }
   // no sched_setaffinity on OSF1
   if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
     printf("# set cpu affinity for core(%d)=%d size=%d failed\n",
             mpi_i, ii, CPU_SETSIZE);
             usleep(100000L);perror("set_affinity");
   }
   usleep(100000); // 100ms, pseudo sync
  }
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  // -- add a dummy loop here for 12s for a taskset call
  if (!mpi_i) {
    printf("# using MPI %d.%d with %d nodes\n", version, subversion, mpi_n);
    printf("# blocksize: 2^%2d = %12lu = %.3e\n", bsize2, bsize, 1.*bsize);
    printf("# mintime[s]:  %2d\n", mintime);
    printf("# pairwise:    %2d\n", pairs);
    printf("# nodeshift:   %2d\n", nshift);
    printf("# align:       %2d\n", align);

    printf("# features: check for data errors at end\n");
  }
  next_src = (int *)malloc(mpi_n*sizeof(int));
  next_dst = (int *)malloc(mpi_n*sizeof(int));
  if (align) {  /* 64 byte = DMA friendly */
    buf1 = (char *)memalign(64,bsize); /* malloc.h */
    buf2 = (char *)memalign(64,bsize);
  } else {
    buf1 = (char *)malloc(bsize); /* replace by numa_alloc_local */
    buf2 = (char *)malloc(bsize);
  }
  if ((!buf1)     || (!buf2)
   || (!next_src) || (!next_dst)) fprintf(stderr,"ERROR: malloc failed");
  /* ToDo: get max error per MPI */
  if ((!buf1)     || (!buf2))     MPI_Abort(MPI_COMM_WORLD, 1);
  if ((!next_src) || (!next_dst)) MPI_Abort(MPI_COMM_WORLD, 2);

  /* prepare order of communication partners (gives flexibility) */
  /* fixed rings   i-1 -> i -> i+1, src!=dst */
  node_dst=(mpi_i+1      )%mpi_n; next_dst[node_dst]=node_dst;
  node_src=(mpi_i-1+mpi_n)%mpi_n; next_src[node_src]=node_src;
  
  /* changing rings i-j -> i -> i+j, src!=dst */
  if (pairs==0 && mpi_n>1)
  for (node_dst=node_src=mpi_i,i2=0; i2<mpi_n; i2++) {
     node_dst=next_dst[node_dst]=(mpi_i+(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
     node_src=next_src[node_src]=(mpi_i-(i2%(mpi_n-1)+1)+mpi_n)%mpi_n;
     if ((i2<5 || i2>mpi_n-3) && !mpi_i) {
       /* print out first 8 and last 4 src/dst-lists */
       printf("# %d:  dst", i2);
       for (i1=0; i1<mpi_n && i1<8; i1++)
         printf(" %2d", (int)(i1+(i2%(mpi_n-1)+1)+mpi_n)%mpi_n);
       if (mpi_n>8) printf(" ...");
       printf("  src");
       for (i1=0; i1<mpi_n && i1<8; i1++)
         printf(" %2d", (int)(i1-(i2%(mpi_n-1)+1)+mpi_n)%mpi_n);
       if (mpi_n>8) printf(" ...");
       printf("\n");
     }
     if (!nshift) break; // no node_shift?
  }

  /* changing pairs i -> n-i+j (fixed pairs: j=1), src==dst */
  if (pairs)
  for (node_dst=node_src=mpi_i,i2=(nshift)?0:1; i2<mpi_n+1; i2++) {
     node_dst=next_dst[node_dst]=(2*mpi_n-mpi_i+i2)%mpi_n;
     node_src=next_src[node_src]=(2*mpi_n-mpi_i+i2)%mpi_n;
     if (i2<8 && !mpi_i) {
       printf("# %d: ", i2);
       for (i1=0; i1<mpi_n && i1<8; i1++)
         printf(" %d", (int)(mpi_n-i1+i2)%mpi_n);
       if (mpi_n>8) printf(" ...");
       printf("\n");
     }
     if (!nshift) break;
  }
  if (!nshift) { /* stop changing, fix! */
    next_dst[node_dst]=node_dst;
    next_src[node_src]=node_src;
  }
  
  /* the last state of node_dst + src is the initial state for the run */

  /* touching and writing memory */
  for (i1=0;i1<bsize;i1++) buf1[i1]=buf2[i1]=(char)i1;

  /* eating the memory to see whether or not libmpi eats vmem at runtime */
  /* this would be worse for spinpack, which tries to fill vmem with data */
  if (minfree_kB!=~0L) {  /* 2014-07-27 memeater */
    char *buf[32]; /* max. 4GB */
    unsigned long mem_eaten=0L;
    int i,j;
    if (!mpi_i)
      printf("# minfree_kB:   %7lu (memeater)\n", minfree_kB);
    /* alloc 8MB blocks until minfree is left */
    buf[0]=malloc(minfree_kB<<10);
    if (!buf[0] && !mpi_i)  /* ToDo: transfer error per mpi */
      printf("WARNING: memeater failed, not enough mem for minfree\n");
    /* eating the rest of memory */
    for (i=31;i>=10;i--) {  /* 2G 1G 512M 256M ... 1kB */
      buf[i]=malloc(1<<i); /* needs minimum 32bit int */
      if (buf[i]) { 
        mem_eaten |= 1<<i;
        for (j=0;j<1<<(i-10);j++) buf[i][j<<10]=j&255; /* touching mem */
      }
    }
    free(buf[0]); /* minfree for mpi routines */
    if (!mpi_i) /* ToDo: get min/max of all threads */
      printf("# mem eaten kB: %7lu\n", mem_eaten>>10);
    if ((mem_eaten>>10)+1==1<<21 && !mpi_i)
      printf("WARNING: memeater failed, please ulimit vmem to max. 2GB\n");
  } /* minfree set */

  MPI_Barrier(MPI_COMM_WORLD);
  if (!ata) {
    if (!mpi_i) printf("# Measure speed of MPI_sendrecv transactions:\n");
    // for (i1=bsize;i1>0;i1>>=1) test_sendrecv(i1); // max to min
    for (i1=bsize0;i1<=bsize;i1<<=1) test_sendrecv(i1); // min to max (2011-10)
  }
  if ( ata) {
    if (!mpi_i) printf("# Measure speed of MPI_alltoall transactions:\n");
    // for (i1=bsize/mpi_n;i1>0;i1>>=1) test_alltoall(i1);
    for (i1=bsize0/mpi_n;i1<=bsize/mpi_n;i1<<=1) test_alltoall(i1);
  }
  
  /* test for correctness of transferred data (bit flips?) */
  /* after lot of loops we dont know about the origin of the error, ...
   * ...  so we just count it and show what kind it is */
  for (i1=0;i1<bsize;i1++) if (buf1[i1]!=(char)i1 || buf2[i1]!=(char)i1)
    {
      if (num_bflips<16)
        printf("ERROR %d: %08lx %02x %02x %02x (bad data)\n",
               mpi_i, i1, buf1[i1], buf2[i1], (char)i1);
      num_bflips++;
    }
  i2= num_bflips;
  MPI_Allreduce(&i2, &num_bflips, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
  if (!mpi_i) printf("# byte-flip errors: %d\n", num_bflips);

  free(buf2);
  free(buf1);  
  MPI_Finalize();
  return 0;
}

/*

  P4-1.3GHz-linux 15.3us      390MB/s(16M)..840MB/s(64K) (2)
  ES45-Tru64-5.1B  2.6us 5%   820MB/s    (2)
                   2.79us    1130MB/s 2% (4)
      overload-2x 26.0us 4%   7.2MB/s 0.2%(8) SR=Sendrecv AA=Alltoall
  GS320-16:    16 12us 860MB/s(128k) (SR) 156us 880MB/s(128k) (AA)  
     runon 0,1  8 12us 409MB/s(128K) (SR)  82us 373MB/s(128k) # psr=0,1,2,3..3
                4  6us 373MB/s(128K) (SR)  23us 440MB/s(64K) (AA)
                2  5us 260MB/s(128K) (SR)  11us 409MB/s(256K) (AA)
  GS1280-32:    2  3.1us 4%   715MB/s 4%  (slow 32CPUs/RAD)
                8  3.9us 3%  1960MB/s(512kB)
                             1560MB/s(1MB) 3% 
               32  6.4us 4%  3930MB/s(32kB)
                             4040MB/s(64kB,128kB)
                             4160MB/s(256kB)
                             3500MB/s(512kB)
                             2500MB/s(1MB) 3%  
  Altix330 numa_alloc_local MPI_via_loopback ToDo: interconnect-topology=star
     # monitor interconnect-traffic: linkstat, dplace?
     MPT-MPI-1.2 (insmod xpmem) MPI_USE_XPMEM=1 MPI_XPMEM_VERBOSE=1 ?
      2 2.09us 1.8GB/s(1M) 1.1GB/s(256K) 1node  numa_alloc_local
      2 1.97us 8G(1M) 17G 15G 13G 12G 8G 5G 3G 2G 1G 300M ... no div psr1,11
      4        1.6GB/s(1M) 1.6GB/s(256K) 4nodes
      8 3.58us 2.5GB/s(1M) 2.7GB/s(256K) 4nodes
     12 4.05us 3.8GB/s(1M) 3.4GB/s(256K) 6nodes numa_alloc_local
     12 4.05us 4.1GB/s(1M) 34GB/s(128K) 6nodes malloc lo (mem remap?)
     12 4.05us 4.3GB/s(1M) 32GB/s(128K) 6nodes malloc xpmem
      # MPI_Sendrecv                    MPI_Alltoall
      4 3.2us  4.2GB/s(1M) 16GB/s(128K) 12.4us  9GB/s(1M) 20GB/s(256K) # running spinpack8
      4 3.1us  4.6GB/s(1M) 20GB/s(128K) 12.4us 11GB/s(1M) 23GB/s(512K) # stopped spinpack8 (parallel-fluent-speed*3.5 14.7.08)
     12 3.9us  3.4GB/s(1M) 37GB/s(128K) 42.0us  6GB/s(1M) 26GB/s(512K) # stopped spinpack8
  2DualAMD265:    2  23us 4%   660MB/s 4% (last!)   baldur
  4Xeon3GHz_tina  4 3.3us 4%   307MB/s 4%  Dec07 node004
    2*1G via .1G  2 160us 5%    15MB/s 3%  Dec07 node4+node012
                  4 191us 4%    28MB/s 3%  Dec07 node4+node012+...
                  8 206us 4%    63MB/s 3%  Dec07 node4+node012+...
  2*4Xeon_1GbE_quantum_parastation_mpich_2008 (4kB/44us) 
      2  lat=0.76us aBW=1511MB/s(256kB) Oct11 shared_mem cset
      8  lat=0.87us aBW=3038MB/s(256kB) Oct11 shared_mem cset
      2  lat=48.9us aBW= 241MB/s(8MB)   Oct11  2*GbE aBW=235MB/s(512kB) cset
     16  lat=48.8us aBW=1660MB/s(16MB)  Oct11 16*GbE aBW=570MB/s(512kB) cset
     16  lat=43.6us aBW=1670MB/s(16MB)  n*104MB/s Oct11 1015MB/s(32kB) noCPUSET
  sicortex.kautz 958*6: 2014-06 clk=500MHz ToDo (340GB/s 2009) max.6hops
     1*2 *  550MB/s             (2 hops, away and back = blocking) -64% TP
     2*1 * 1.53GB/s =  3.1GB/s 1.9us (128MB) throughput = BW  (3x multi rail)
     2*6 *  218MB/s =  2.6GB/s 2.4us (16M-256MB)   2013-05
    12*1 ...                  throughput reduced
    81*1 *  730MB/s =   59GB/s ...                  -52% blocking net
   243*1 *  500MB/s =  120GB/s 2.6us                -67% blocking net
   958*1 *  316MB/s =  303GB/s 3.7us nur 256 loops! -80% blocking net
   958*6 *   48MB/s =  280GB/s 8.3us nur 256 loops! 2013-05  ca. 100*IB-QDR
  sicortex.asgard 12*6 2009 clk=680MHz mpich2 3pair-distances: 1.22,1.24,1.27us
     2*1 * 1.72GB/s =  3.4GB/s 1.3us (64MB) 2009 (speed2014 + 1..3%)
    12*6 *  213MB/s = 15.4GB/s 2.4us (64MB)
     2*1 * 1.52GB/s =  3.0GB/s 1.3us (16M-128M)  2014-06 throughput
     2*6 *  240MB/s =  2.9GB/s 1.9us (64M-128MB)  -0% blocking net
    12*1 * 1.41GB/s = 17.0GB/s 1.3us (16M-128MB)  -7% blocking net
    12*6 *  216MB/s = 15.6GB/s 2.5us (2M-128MB)  -10% blocking net
  sicortex.mkautz(2009) 12*6 2014-06 clk=500MHz max.3hops
     2*1 * 1.45GB/s =  2.9GB/s 1.8us (16M-128MB)      multi-rail-bonus! (-0%)
     2*6 *  220MB/s =  2.6GB/s 2.4us  (1M-128MB, 80s) 5% multi-rail-bonus!
    12*1 * 1.36GB/s = 16.3GB/s 1.8us (16M-256MB, 24s)  -7% blocking net
    12*6 *  206MB/s = 14.8GB/s 3.0us  (4M-128MB, 600s) -7% blocking net
    # may crash on 6*2*256MB=3GB of 3.6GB without any msg!
     
  ToDo: show scaling speed(threads)

*/
